@@ -25,7 +25,7 @@
 #include "shrpx_http2_downstream_connection.h"
 
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>
+#  include <unistd.h>
 #endif // HAVE_UNISTD_H
 
 #include "http-parser/http_parser.h"
@@ -37,17 +37,21 @@
 #include "shrpx_error.h"
 #include "shrpx_http.h"
 #include "shrpx_http2_session.h"
+#include "shrpx_worker.h"
+#include "shrpx_log.h"
 #include "http2.h"
 #include "util.h"
+#include "ssl_compat.h"
 
 using namespace nghttp2;
 
 namespace shrpx {
 
-Http2DownstreamConnection::Http2DownstreamConnection(
-    DownstreamConnectionPool *dconn_pool, Http2Session *http2session)
-    : DownstreamConnection(dconn_pool), dlnext(nullptr), dlprev(nullptr),
-      http2session_(http2session), sd_(nullptr) {}
+Http2DownstreamConnection::Http2DownstreamConnection(Http2Session *http2session)
+    : dlnext(nullptr),
+      dlprev(nullptr),
+      http2session_(http2session),
+      sd_(nullptr) {}
 
 Http2DownstreamConnection::~Http2DownstreamConnection() {
   if (LOG_ENABLED(INFO)) {
@@ -58,23 +62,25 @@ Http2DownstreamConnection::~Http2DownstreamConnection() {
     downstream_->disable_downstream_wtimer();
 
     uint32_t error_code;
-    if (downstream_->get_request_state() == Downstream::STREAM_CLOSED &&
+    if (downstream_->get_request_state() == DownstreamState::STREAM_CLOSED &&
         downstream_->get_upgraded()) {
       // For upgraded connection, send NO_ERROR.  Should we consider
-      // request states other than Downstream::STREAM_CLOSED ?
+      // request states other than DownstreamState::STREAM_CLOSED ?
       error_code = NGHTTP2_NO_ERROR;
     } else {
       error_code = NGHTTP2_INTERNAL_ERROR;
     }
 
-    if (http2session_->get_state() == Http2Session::CONNECTED &&
+    if (http2session_->get_state() == Http2SessionState::CONNECTED &&
         downstream_->get_downstream_stream_id() != -1) {
       submit_rst_stream(downstream_, error_code);
 
-      http2session_->consume(downstream_->get_downstream_stream_id(),
-                             downstream_->get_response_datalen());
+      auto &resp = downstream_->response();
 
-      downstream_->reset_response_datalen();
+      http2session_->consume(downstream_->get_downstream_stream_id(),
+                             resp.unconsumed_body_length);
+
+      resp.unconsumed_body_length = 0;
 
       http2session_->signal_write();
     }
@@ -91,13 +97,18 @@ int Http2DownstreamConnection::attach_downstream(Downstream *downstream) {
     DCLOG(INFO, this) << "Attaching to DOWNSTREAM:" << downstream;
   }
   http2session_->add_downstream_connection(this);
-  if (http2session_->get_state() == Http2Session::DISCONNECTED) {
-    http2session_->signal_write();
-  }
+  http2session_->signal_write();
 
   downstream_ = downstream;
   downstream_->reset_downstream_rtimer();
 
+  auto &req = downstream_->request();
+
+  // HTTP/2 disables HTTP Upgrade.
+  if (req.method != HTTP_CONNECT && req.connect_proto == ConnectProto::NONE) {
+    req.upgrade_request = false;
+  }
+
   return 0;
 }
 
@@ -105,15 +116,18 @@ void Http2DownstreamConnection::detach_downstream(Downstream *downstream) {
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, this) << "Detaching from DOWNSTREAM:" << downstream;
   }
-  if (submit_rst_stream(downstream) == 0) {
-    http2session_->signal_write();
-  }
+
+  auto &resp = downstream_->response();
 
   if (downstream_->get_downstream_stream_id() != -1) {
-    http2session_->consume(downstream_->get_downstream_stream_id(),
-                           downstream_->get_response_datalen());
+    if (submit_rst_stream(downstream) == 0) {
+      http2session_->signal_write();
+    }
 
-    downstream_->reset_response_datalen();
+    http2session_->consume(downstream_->get_downstream_stream_id(),
+                           resp.unconsumed_body_length);
+
+    resp.unconsumed_body_length = 0;
 
     http2session_->signal_write();
   }
@@ -126,12 +140,12 @@ void Http2DownstreamConnection::detach_downstream(Downstream *downstream) {
 int Http2DownstreamConnection::submit_rst_stream(Downstream *downstream,
                                                  uint32_t error_code) {
   int rv = -1;
-  if (http2session_->get_state() == Http2Session::CONNECTED &&
+  if (http2session_->get_state() == Http2SessionState::CONNECTED &&
       downstream->get_downstream_stream_id() != -1) {
     switch (downstream->get_response_state()) {
-    case Downstream::MSG_RESET:
-    case Downstream::MSG_BAD_HEADER:
-    case Downstream::MSG_COMPLETE:
+    case DownstreamState::MSG_RESET:
+    case DownstreamState::MSG_BAD_HEADER:
+    case DownstreamState::MSG_COMPLETE:
       break;
     default:
       if (LOG_ENABLED(INFO)) {
@@ -158,51 +172,37 @@ ssize_t http2_data_read_callback(nghttp2_session *session, int32_t stream_id,
   if (!sd || !sd->dconn) {
     return NGHTTP2_ERR_DEFERRED;
   }
-  auto dconn = static_cast<Http2DownstreamConnection *>(source->ptr);
+  auto dconn = sd->dconn;
   auto downstream = dconn->get_downstream();
   if (!downstream) {
     // In this case, RST_STREAM should have been issued. But depending
     // on the priority, DATA frame may come first.
     return NGHTTP2_ERR_DEFERRED;
   }
+  const auto &req = downstream->request();
   auto input = downstream->get_request_buf();
-  auto nread = input->remove(buf, length);
-  auto input_empty = input->rleft() == 0;
 
-  if (nread > 0) {
-    // This is important because it will handle flow control
-    // stuff.
-    if (downstream->get_upstream()->resume_read(SHRPX_NO_BUFFER, downstream,
-                                                nread) != 0) {
-      // In this case, downstream may be deleted.
-      return NGHTTP2_ERR_CALLBACK_FAILURE;
-    }
+  auto nread = std::min(input->rleft(), length);
+  auto input_empty = input->rleft() == nread;
 
-    // Check dconn is still alive because Upstream::resume_read()
-    // may delete downstream which will delete dconn.
-    if (sd->dconn == nullptr) {
-      return NGHTTP2_ERR_DEFERRED;
-    }
-  }
+  *data_flags |= NGHTTP2_DATA_FLAG_NO_COPY;
 
   if (input_empty &&
-      downstream->get_request_state() == Downstream::MSG_COMPLETE &&
+      downstream->get_request_state() == DownstreamState::MSG_COMPLETE &&
       // If connection is upgraded, don't set EOF flag, since HTTP/1
       // will set MSG_COMPLETE to request state after upgrade response
       // header is seen.
-      (!downstream->get_upgrade_request() ||
-       (downstream->get_response_state() == Downstream::HEADER_COMPLETE &&
+      (!req.upgrade_request ||
+       (downstream->get_response_state() == DownstreamState::HEADER_COMPLETE &&
         !downstream->get_upgraded()))) {
 
     *data_flags |= NGHTTP2_DATA_FLAG_EOF;
 
-    auto &trailers = downstream->get_request_trailers();
+    const auto &trailers = req.fs.trailers();
     if (!trailers.empty()) {
       std::vector<nghttp2_nv> nva;
       nva.reserve(trailers.size());
-      // We cannot use nocopy version, since nva may be touched after
-      // Downstream object is deleted.
-      http2::copy_headers_to_nva(nva, trailers);
+      http2::copy_headers_to_nva_nocopy(nva, trailers, http2::HDOP_STRIP_ALL);
       if (!nva.empty()) {
         rv = nghttp2_submit_trailer(session, stream_id, nva.data(), nva.size());
         if (rv != 0) {
@@ -216,12 +216,6 @@ ssize_t http2_data_read_callback(nghttp2_session *session, int32_t stream_id,
     }
   }
 
-  if (!input_empty) {
-    downstream->reset_downstream_wtimer();
-  } else {
-    downstream->disable_downstream_wtimer();
-  }
-
   if (nread == 0 && (*data_flags & NGHTTP2_DATA_FLAG_EOF) == 0) {
     downstream->disable_downstream_wtimer();
 
@@ -237,7 +231,7 @@ int Http2DownstreamConnection::push_request_headers() {
   if (!downstream_) {
     return 0;
   }
-  if (!http2session_->can_push_request()) {
+  if (!http2session_->can_push_request(downstream_)) {
     // The HTTP2 session to the backend has not been established or
     // connection is now being checked.  This function will be called
     // again just after it is established.
@@ -248,162 +242,256 @@ int Http2DownstreamConnection::push_request_headers() {
 
   downstream_->set_request_pending(false);
 
-  auto method = downstream_->get_request_method();
-  auto no_host_rewrite = get_config()->no_host_rewrite ||
-                         get_config()->http2_proxy ||
-                         get_config()->client_proxy || method == HTTP_CONNECT;
+  const auto &req = downstream_->request();
+
+  if (req.connect_proto != ConnectProto::NONE &&
+      !http2session_->get_allow_connect_proto()) {
+    return -1;
+  }
+
+  auto &balloc = downstream_->get_block_allocator();
+
+  auto config = get_config();
+  auto &httpconf = config->http;
+  auto &http2conf = config->http2;
+
+  auto no_host_rewrite = httpconf.no_host_rewrite || config->http2_proxy ||
+                         req.regular_connect_method();
 
   // http2session_ has already in CONNECTED state, so we can get
   // addr_idx here.
-  auto addr_idx = http2session_->get_addr_idx();
-  auto group = http2session_->get_group();
-  auto downstream_hostport = get_config()
-                                 ->downstream_addr_groups[group]
-                                 .addrs[addr_idx]
-                                 .hostport.get();
+  const auto &downstream_hostport = http2session_->get_addr()->hostport;
 
   // For HTTP/1.0 request, there is no authority in request.  In that
   // case, we use backend server's host nonetheless.
-  const char *authority = downstream_hostport;
-  auto &req_authority = downstream_->get_request_http2_authority();
-  if (no_host_rewrite && !req_authority.empty()) {
-    authority = req_authority.c_str();
+  auto authority = StringRef(downstream_hostport);
+
+  if (no_host_rewrite && !req.authority.empty()) {
+    authority = req.authority;
   }
 
   downstream_->set_request_downstream_host(authority);
 
-  auto nheader = downstream_->get_request_headers().size();
-
   size_t num_cookies = 0;
-  if (!get_config()->http2_no_cookie_crumbling) {
+  if (!http2conf.no_cookie_crumbling) {
     num_cookies = downstream_->count_crumble_request_cookie();
   }
 
-  // 8 means:
+  // 11 means:
   // 1. :method
   // 2. :scheme
   // 3. :path
-  // 4. :authority
-  // 5. via (optional)
-  // 6. x-forwarded-for (optional)
-  // 7. x-forwarded-proto (optional)
-  // 8. te (optional)
+  // 4. :authority (or host)
+  // 5. :protocol (optional)
+  // 6. via (optional)
+  // 7. x-forwarded-for (optional)
+  // 8. x-forwarded-proto (optional)
+  // 9. te (optional)
+  // 10. forwarded (optional)
+  // 11. early-data (optional)
   auto nva = std::vector<nghttp2_nv>();
-  nva.reserve(nheader + 8 + num_cookies +
-              get_config()->add_request_headers.size());
+  nva.reserve(req.fs.headers().size() + 11 + num_cookies +
+              httpconf.add_request_headers.size());
 
-  nva.push_back(
-      http2::make_nv_lc_nocopy(":method", http2::to_method_string(method)));
+  if (req.connect_proto == ConnectProto::WEBSOCKET) {
+    nva.push_back(http2::make_nv_ll(":method", "CONNECT"));
+    nva.push_back(http2::make_nv_ll(":protocol", "websocket"));
+  } else {
+    nva.push_back(http2::make_nv_ls_nocopy(
+        ":method", http2::to_method_string(req.method)));
+  }
 
-  auto &scheme = downstream_->get_request_http2_scheme();
+  if (!req.regular_connect_method()) {
+    assert(!req.scheme.empty());
 
-  nva.push_back(http2::make_nv_lc_nocopy(":authority", authority));
+    auto addr = http2session_->get_addr();
+    assert(addr);
+    // We will handle more protocol scheme upgrade in the future.
+    if (addr->tls && addr->upgrade_scheme && req.scheme == "http") {
+      nva.push_back(http2::make_nv_ll(":scheme", "https"));
+    } else {
+      nva.push_back(http2::make_nv_ls_nocopy(":scheme", req.scheme));
+    }
 
-  if (method != HTTP_CONNECT) {
-    assert(!scheme.empty());
-
-    nva.push_back(http2::make_nv_ls_nocopy(":scheme", scheme));
-
-    auto &path = downstream_->get_request_path();
-    if (method == HTTP_OPTIONS && path.empty()) {
+    if (req.method == HTTP_OPTIONS && req.path.empty()) {
       nva.push_back(http2::make_nv_ll(":path", "*"));
     } else {
-      nva.push_back(http2::make_nv_ls_nocopy(":path", path));
+      nva.push_back(http2::make_nv_ls_nocopy(":path", req.path));
     }
+
+    if (!req.no_authority || req.connect_proto != ConnectProto::NONE) {
+      nva.push_back(http2::make_nv_ls_nocopy(":authority", authority));
+    } else {
+      nva.push_back(http2::make_nv_ls_nocopy("host", authority));
+    }
+  } else {
+    nva.push_back(http2::make_nv_ls_nocopy(":authority", authority));
   }
 
-  http2::copy_headers_to_nva_nocopy(nva, downstream_->get_request_headers());
+  auto &fwdconf = httpconf.forwarded;
+  auto &xffconf = httpconf.xff;
+  auto &xfpconf = httpconf.xfp;
+  auto &earlydataconf = httpconf.early_data;
 
-  bool chunked_encoding = false;
-  auto transfer_encoding =
-      downstream_->get_request_header(http2::HD_TRANSFER_ENCODING);
-  if (transfer_encoding &&
-      util::strieq_l("chunked", (*transfer_encoding).value)) {
-    chunked_encoding = true;
-  }
+  uint32_t build_flags =
+      (fwdconf.strip_incoming ? http2::HDOP_STRIP_FORWARDED : 0) |
+      (xffconf.strip_incoming ? http2::HDOP_STRIP_X_FORWARDED_FOR : 0) |
+      (xfpconf.strip_incoming ? http2::HDOP_STRIP_X_FORWARDED_PROTO : 0) |
+      (earlydataconf.strip_incoming ? http2::HDOP_STRIP_EARLY_DATA : 0) |
+      http2::HDOP_STRIP_SEC_WEBSOCKET_KEY;
 
-  if (!get_config()->http2_no_cookie_crumbling) {
+  http2::copy_headers_to_nva_nocopy(nva, req.fs.headers(), build_flags);
+
+  if (!http2conf.no_cookie_crumbling) {
     downstream_->crumble_request_cookie(nva);
   }
 
-  std::string xff_value;
-  auto xff = downstream_->get_request_header(http2::HD_X_FORWARDED_FOR);
-  if (get_config()->add_x_forwarded_for) {
-    if (xff && !get_config()->strip_incoming_x_forwarded_for) {
-      xff_value = (*xff).value;
-      xff_value += ", ";
+  auto upstream = downstream_->get_upstream();
+  auto handler = upstream->get_client_handler();
+
+#if OPENSSL_1_1_1_API
+  auto conn = handler->get_connection();
+
+  if (conn->tls.ssl && !SSL_is_init_finished(conn->tls.ssl)) {
+    nva.push_back(http2::make_nv_ll("early-data", "1"));
+  }
+#endif // OPENSSL_1_1_1_API
+
+  auto fwd =
+      fwdconf.strip_incoming ? nullptr : req.fs.header(http2::HD_FORWARDED);
+
+  if (fwdconf.params) {
+    auto params = fwdconf.params;
+
+    if (config->http2_proxy || req.regular_connect_method()) {
+      params &= ~FORWARDED_PROTO;
     }
-    xff_value +=
-        downstream_->get_upstream()->get_client_handler()->get_ipaddr();
-    nva.push_back(http2::make_nv_ls("x-forwarded-for", xff_value));
-  } else if (xff && !get_config()->strip_incoming_x_forwarded_for) {
-    nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-for", (*xff).value));
+
+    auto value = http::create_forwarded(
+        balloc, params, handler->get_forwarded_by(),
+        handler->get_forwarded_for(), req.authority, req.scheme);
+
+    if (fwd || !value.empty()) {
+      if (fwd) {
+        if (value.empty()) {
+          value = fwd->value;
+        } else {
+          value = concat_string_ref(balloc, fwd->value,
+                                    StringRef::from_lit(", "), value);
+        }
+      }
+
+      nva.push_back(http2::make_nv_ls_nocopy("forwarded", value));
+    }
+  } else if (fwd) {
+    nva.push_back(http2::make_nv_ls_nocopy("forwarded", fwd->value));
   }
 
-  if (!get_config()->http2_proxy && !get_config()->client_proxy &&
-      downstream_->get_request_method() != HTTP_CONNECT) {
-    // We use same protocol with :scheme header field
-    nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-proto", scheme));
+  auto xff = xffconf.strip_incoming ? nullptr
+                                    : req.fs.header(http2::HD_X_FORWARDED_FOR);
+
+  if (xffconf.add) {
+    StringRef xff_value;
+    const auto &addr = upstream->get_client_handler()->get_ipaddr();
+    if (xff) {
+      xff_value = concat_string_ref(balloc, xff->value,
+                                    StringRef::from_lit(", "), addr);
+    } else {
+      xff_value = addr;
+    }
+    nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-for", xff_value));
+  } else if (xff) {
+    nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-for", xff->value));
   }
 
-  std::string via_value;
-  auto via = downstream_->get_request_header(http2::HD_VIA);
-  if (get_config()->no_via) {
+  if (!config->http2_proxy && !req.regular_connect_method()) {
+    auto xfp = xfpconf.strip_incoming
+                   ? nullptr
+                   : req.fs.header(http2::HD_X_FORWARDED_PROTO);
+
+    if (xfpconf.add) {
+      StringRef xfp_value;
+      // We use same protocol with :scheme header field
+      if (xfp) {
+        xfp_value = concat_string_ref(balloc, xfp->value,
+                                      StringRef::from_lit(", "), req.scheme);
+      } else {
+        xfp_value = req.scheme;
+      }
+      nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-proto", xfp_value));
+    } else if (xfp) {
+      nva.push_back(http2::make_nv_ls_nocopy("x-forwarded-proto", xfp->value));
+    }
+  }
+
+  auto via = req.fs.header(http2::HD_VIA);
+  if (httpconf.no_via) {
     if (via) {
       nva.push_back(http2::make_nv_ls_nocopy("via", (*via).value));
     }
   } else {
+    size_t vialen = 16;
     if (via) {
-      via_value = (*via).value;
-      via_value += ", ";
+      vialen += via->value.size() + 2;
     }
-    via_value += http::create_via_header_value(
-        downstream_->get_request_major(), downstream_->get_request_minor());
-    nva.push_back(http2::make_nv_ls("via", via_value));
+
+    auto iov = make_byte_ref(balloc, vialen + 1);
+    auto p = iov.base;
+
+    if (via) {
+      p = std::copy(std::begin(via->value), std::end(via->value), p);
+      p = util::copy_lit(p, ", ");
+    }
+    p = http::create_via_header_value(p, req.http_major, req.http_minor);
+    *p = '\0';
+
+    nva.push_back(http2::make_nv_ls_nocopy("via", StringRef{iov.base, p}));
   }
 
-  auto te = downstream_->get_request_header(http2::HD_TE);
+  auto te = req.fs.header(http2::HD_TE);
   // HTTP/1 upstream request can contain keyword other than
   // "trailers".  We just forward "trailers".
   // TODO more strict handling required here.
-  if (te && util::strifind(te->value.c_str(), "trailers")) {
+  if (te && http2::contains_trailers(te->value)) {
     nva.push_back(http2::make_nv_ll("te", "trailers"));
   }
 
-  for (auto &p : get_config()->add_request_headers) {
-    nva.push_back(http2::make_nv_nocopy(p.first, p.second));
+  for (auto &p : httpconf.add_request_headers) {
+    nva.push_back(http2::make_nv_nocopy(p.name, p.value));
   }
 
   if (LOG_ENABLED(INFO)) {
     std::stringstream ss;
     for (auto &nv : nva) {
-      ss << TTY_HTTP_HD << nv.name << TTY_RST << ": " << nv.value << "\n";
+      ss << TTY_HTTP_HD << StringRef{nv.name, nv.namelen} << TTY_RST << ": "
+         << StringRef{nv.value, nv.valuelen} << "\n";
     }
     DCLOG(INFO, this) << "HTTP request headers\n" << ss.str();
   }
 
-  auto content_length =
-      downstream_->get_request_header(http2::HD_CONTENT_LENGTH);
-  // TODO check content-length: 0 case
+  auto transfer_encoding = req.fs.header(http2::HD_TRANSFER_ENCODING);
 
-  if (downstream_->get_request_method() == HTTP_CONNECT || chunked_encoding ||
-      content_length || downstream_->get_request_http2_expect_body()) {
+  nghttp2_data_provider *data_prdptr = nullptr;
+  nghttp2_data_provider data_prd;
+
+  // Add body as long as transfer-encoding is given even if
+  // req.fs.content_length == 0 to forward trailer fields.
+  if (req.method == HTTP_CONNECT || req.connect_proto != ConnectProto::NONE ||
+      transfer_encoding || req.fs.content_length > 0 || req.http2_expect_body) {
     // Request-body is expected.
-    nghttp2_data_provider data_prd;
-    data_prd.source.ptr = this;
-    data_prd.read_callback = http2_data_read_callback;
-    rv = http2session_->submit_request(this, downstream_->get_priority(),
-                                       nva.data(), nva.size(), &data_prd);
-  } else {
-    rv = http2session_->submit_request(this, downstream_->get_priority(),
-                                       nva.data(), nva.size(), nullptr);
+    data_prd = {{}, http2_data_read_callback};
+    data_prdptr = &data_prd;
   }
+
+  rv = http2session_->submit_request(this, nva.data(), nva.size(), data_prdptr);
   if (rv != 0) {
     DCLOG(FATAL, this) << "nghttp2_submit_request() failed";
     return -1;
   }
 
-  downstream_->reset_downstream_wtimer();
+  if (data_prdptr) {
+    downstream_->reset_downstream_wtimer();
+  }
 
   http2session_->signal_write();
   return 0;
@@ -411,6 +499,14 @@ int Http2DownstreamConnection::push_request_headers() {
 
 int Http2DownstreamConnection::push_upload_data_chunk(const uint8_t *data,
                                                       size_t datalen) {
+  if (!downstream_->get_request_header_sent()) {
+    auto output = downstream_->get_blocked_request_buf();
+    auto &req = downstream_->request();
+    output->append(data, datalen);
+    req.unconsumed_body_length += datalen;
+    return 0;
+  }
+
   int rv;
   auto output = downstream_->get_request_buf();
   output->append(data, datalen);
@@ -428,6 +524,11 @@ int Http2DownstreamConnection::push_upload_data_chunk(const uint8_t *data,
 }
 
 int Http2DownstreamConnection::end_upload_data() {
+  if (!downstream_->get_request_header_sent()) {
+    downstream_->set_blocked_request_data_eof(true);
+    return 0;
+  }
+
   int rv;
   if (downstream_->get_downstream_stream_id() != -1) {
     rv = http2session_->resume_data(this);
@@ -446,8 +547,7 @@ int Http2DownstreamConnection::resume_read(IOCtrlReason reason,
                                            size_t consumed) {
   int rv;
 
-  if (http2session_->get_state() != Http2Session::CONNECTED ||
-      !http2session_->get_flow_control()) {
+  if (http2session_->get_state() != Http2SessionState::CONNECTED) {
     return 0;
   }
 
@@ -456,8 +556,6 @@ int Http2DownstreamConnection::resume_read(IOCtrlReason reason,
   }
 
   if (consumed > 0) {
-    assert(downstream_->get_response_datalen() >= consumed);
-
     rv = http2session_->consume(downstream_->get_downstream_stream_id(),
                                 consumed);
 
@@ -465,7 +563,9 @@ int Http2DownstreamConnection::resume_read(IOCtrlReason reason,
       return -1;
     }
 
-    downstream_->dec_response_datalen(consumed);
+    auto &resp = downstream_->response();
+
+    resp.unconsumed_body_length -= consumed;
 
     http2session_->signal_write();
   }
@@ -498,24 +598,6 @@ StreamData *Http2DownstreamConnection::detach_stream_data() {
   return nullptr;
 }
 
-int Http2DownstreamConnection::on_priority_change(int32_t pri) {
-  int rv;
-  if (downstream_->get_priority() == pri) {
-    return 0;
-  }
-  downstream_->set_priority(pri);
-  if (http2session_->get_state() != Http2Session::CONNECTED) {
-    return 0;
-  }
-  rv = http2session_->submit_priority(this, pri);
-  if (rv != 0) {
-    DLOG(FATAL, this) << "nghttp2_submit_priority() failed";
-    return -1;
-  }
-  http2session_->signal_write();
-  return 0;
-}
-
 int Http2DownstreamConnection::on_timeout() {
   if (!downstream_) {
     return 0;
@@ -524,10 +606,11 @@ int Http2DownstreamConnection::on_timeout() {
   return submit_rst_stream(downstream_, NGHTTP2_NO_ERROR);
 }
 
-size_t Http2DownstreamConnection::get_group() const {
-  // HTTP/2 backend connections are managed by Http2Session object,
-  // and it stores group index.
-  return http2session_->get_group();
+const std::shared_ptr<DownstreamAddrGroup> &
+Http2DownstreamConnection::get_downstream_addr_group() const {
+  return http2session_->get_downstream_addr_group();
 }
 
+DownstreamAddr *Http2DownstreamConnection::get_addr() const { return nullptr; }
+
 } // namespace shrpx
